
//#define ENTROPYLOG

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <SDL/SDL.h>
#include <time.h>
#ifdef ENTROPYLOG
	#include <zlib.h>
	#include <zconf.h>
#endif
#include "twister.h"
#include "mwcrand.h"
#include "draw.h"
#include "fungus.h"

#define VIDEOFLAGS  SDL_SWSURFACE | SDL_DOUBLEBUF | SDL_RESIZABLE
#define ERASESIZE		20

MTRand mtr;
MWCRand mwcr;

// ms per frame. no semaphore.
Uint32 fps_interval = 0;

Uint32 WaitUnlock (Uint32 interval, void * semaphore)
{
  if(SDL_SemValue( (SDL_sem*)semaphore) == 0 ){
    SDL_SemPost( (SDL_sem*)semaphore );
    SDL_Event evnt;
    evnt.type = SDL_USEREVENT;
    SDL_PushEvent( &evnt ); // wake up event loop
  }
  if( fps_interval != 0 )
  	return fps_interval;
  else
  	return 50;
}


void splash ( SDL_Surface * screen, std::string & filename )
{
  SDL_FillRect( screen, NULL, C_BG );
  // try to load startup splash
  SDL_Surface * imgfile = SDL_LoadBMP( filename.c_str() );
  if( imgfile ){
  	SDL_Surface * img = SDL_DisplayFormat( imgfile );
  	SDL_FreeSurface( imgfile );
  	SDL_Rect src;
  	src.x = src.y = 0;
  	src.w = img->w;
  	src.h = img->h;
  	SDL_Rect dst;
  	dst.x = ( screen->w - img->w ) / 2;
  	dst.y = ( screen->h - img->h ) / 2;
  	dst.w = img->w;
  	dst.h = img->h;
  	SDL_BlitSurface( img, &src, screen, &dst );
  	SDL_FreeSurface( img );
  }
}


int main (int argc, char *argv[])
{
  int w = 400, h = 400;
  std::string splashfile = "splash.bmp";
  if( argc > 1 ){ // maybe file specified on command line?
  	FILE * temp = fopen( argv[1], "r" );
  	if( temp ){
  		fclose( temp );
  		splashfile = argv[1];
  	}
  }
  SDL_Surface * spimg = SDL_LoadBMP( splashfile.c_str() );
  if( spimg ){
  	w = spimg->w;
  	h = spimg->h;
  	SDL_FreeSurface( spimg );
  }

  /////////////////////////////////////////// sys init
  if(SDL_Init( SDL_INIT_TIMER | SDL_INIT_VIDEO) < 0 )
    exit(1);
  atexit( SDL_Quit );
  SDL_Surface * screen = SDL_SetVideoMode( w, h, 32, VIDEOFLAGS );
  if( screen == NULL )
    exit(2);
  SDL_WM_SetCaption( "numberfungus", NULL );
  ////////////////////////////////////////////////////

  splash( screen, splashfile );
  SDL_Flip( screen );

  // set up timer
  SDL_sem * sem_fpslimit = SDL_CreateSemaphore(0);
  SDL_TimerID tmr_fpslimit = SDL_AddTimer( fps_interval, &WaitUnlock, sem_fpslimit );

  bool done = false;
  bool flip = false;
#ifdef ENTROPYLOG
  unsigned long frames = 0;
  FILE * entlog = fopen( "log.csv", "w" );
#endif
  while(!done){

  	// get mouse
  	int mx, my;
  	Uint8 mb = SDL_GetMouseState( &mx, &my );

  	// handle events
    SDL_Event event;
    while( SDL_PollEvent(&event) )
      switch( event.type ){

        case SDL_KEYDOWN:
          switch( event.key.keysym.sym ){

            case SDLK_ESCAPE:
              splash( screen, splashfile );
              flip = true;
              break;

            case SDLK_0:	fps_interval = 0;		break;
            case SDLK_1:	fps_interval = 2000;		break;
            case SDLK_2:	fps_interval = 1000;		break;
            case SDLK_3:	fps_interval = 500;		break;
            case SDLK_4:	fps_interval = 200;		break;
            case SDLK_5:	fps_interval = 100;		break;
            case SDLK_6:	fps_interval = 50;		break;
            case SDLK_7:	fps_interval = 30;		break;
            case SDLK_8:	fps_interval = 10;		break;
            case SDLK_9:	fps_interval = 1;		break;

            default:
            	break;
          }
          break;

        case SDL_QUIT:
          done = true;
          break;

        case SDL_MOUSEBUTTONDOWN:
        case SDL_MOUSEMOTION:
          if(mb & SDL_BUTTON(1)){
              SDL_LockSurface(screen);
              	//Uint32 c = mtr.randBits(8) % (mtr.randBits(8) | 1);
              	//pset( screen, mx, my, c | c << 8 | c << 16 | c << 24 );//mtr.randInt() | 0x07070707 );
              	pset( screen, mx, my, mtr.randInt() );
              SDL_UnlockSurface( screen );
              flip = true;
            }
          break;

        case SDL_VIDEORESIZE:
          SDL_FreeSurface( screen );
          w = event.resize.w;
          h = event.resize.h;
          screen = SDL_SetVideoMode( w, h, 32, VIDEOFLAGS );
          if( screen == NULL )
            exit(2);
          splash( screen, splashfile );
          flip = true;
          break;
      }
    // end PollEvent

    if(mb & SDL_BUTTON(3)){
    	SDL_LockSurface( screen );
  			circle( screen, mx, my, -ERASESIZE, C_BG );
  		SDL_UnlockSurface( screen );
  	  flip = true;
    }

if(fps_interval == 1 || SDL_SemTryWait(sem_fpslimit) == 0){ // turbo shortcut
///////////////////////////////////////////////////////////////////////////
if( fps_interval != 0 ){
  mwcr.seed( mtr.randInt() );
  SDL_LockSurface( screen );
  	decay( screen, ((w * h) / 3200) );
  	for(int mult = 0; mult < 10; mult++){
  		fungus( screen, ((w * h) / 80) );
  	}

#ifdef ENTROPYLOG
  	frames++;
  	if( frames % 50 == 0 ){
  		uLongf ssize = screen->pitch * screen->h;
  		uLongf dsize = ssize + 1024;
  		Bytef * dbuf = new Bytef[ dsize ];
  		compress( dbuf, &dsize, (Bytef*)screen->pixels, ssize );
  		delete[] dbuf;
  		fprintf( entlog, "%u, %u\n", frames, dsize );
  	}
#endif
  SDL_UnlockSurface( screen );
  flip = true;
}
///////////////////////////////////////////////////////////////////////////
	if( flip ){
		SDL_UpdateRect( screen, 0,0,0,0 );
		SDL_Flip( screen );
		flip = false;
	}
}
 	else
    SDL_WaitEvent( NULL );

  } // !done
#ifdef ENTROPYLOG
  fclose( entlog );
#endif
  SDL_RemoveTimer( tmr_fpslimit );
  SDL_DestroySemaphore( sem_fpslimit );
  SDL_FreeSurface( screen );
  SDL_Quit();
  return 0;
}
